import xlwt
from openpyxl.styles import PatternFill
from openpyxl.styles.colors import COLOR_INDEX
from RiceXL.RiceUtils import get_color_rgb, get_color_rgb_by_index

FILL_TYPE = (
    'none', #0
    'solid', #1
    'mediumGray', #2
    'darkGray', #3
    'lightGray', #4
    'darkHorizontal', #5
    'darkVertical', #6
    'darkDown', #7
    'darkUp', #8
    'darkGrid', #9
    'darkTrellis', #10
    'lightHorizontal', #11
    'lightVertical', #12
    'lightDown', #13
    'lightUp', #14
    'lightGrid', #15
    'lightTrellis', #16
    'gray125', #17
    'gray0625' #18
)

class RicePattern():
    _type = None
    _start_color = None
    _end_color = None

    @property
    def type(self):
        return self._type
    
    @type.setter
    def type(self, type):
        self._type = type
        self.pattern_xls.pattern = FILL_TYPE.index(self._type)
        self.pattern_xlsx.fill_type = None if self._type == 'none' else self._type
    
    @property
    def start_color(self):
        return self._start_color

    @start_color.setter
    def start_color(self, color):
        self._start_color = color
        self.pattern_xls.pattern_fore_colour = COLOR_INDEX.index(color)
        self.pattern_xlsx.start_color = color
    
    @property
    def end_color(self):
        return self._end_color
    
    @end_color.setter
    def end_color(self, color):
        self._end_color = color
        self.pattern_xls.pattern_back_colour = COLOR_INDEX.index(color)
        self.pattern_xlsx.end_color = color
    
    def init_xls(self, pattern):
        rice_pattern = RicePattern()
        rice_pattern.type = FILL_TYPE[pattern.pattern]
        rice_pattern.start_color = get_color_rgb_by_index(pattern.pattern_fore_colour)
        rice_pattern.end_color = get_color_rgb_by_index(pattern.pattern_back_colour)
        return rice_pattern
    
    def init_xlsx(self, pattern):
        rice_pattern = RicePattern()
        rice_pattern.type = pattern.fill_type if pattern.fill_type is not None else 'none'
        rice_pattern.start_color = get_color_rgb(pattern.start_color)
        rice_pattern.end_color = get_color_rgb(pattern.end_color)
        return rice_pattern

    def __init__(self):
        self.pattern_xls = xlwt.Pattern()
        self.pattern_xlsx = PatternFill()


import re
from openpyxl.styles.colors import COLOR_INDEX
from openpyxl.utils import get_column_letter
from openpyxl.utils import column_index_from_string

def get_color_rgb(color):
    # Only support rgb
    if color.type == 'rgb':
        color = '00' + color.rgb[2:]
    else:
        color = '00000000'
    return color

def get_color_index_by_rgb(color):
    return COLOR_INDEX.index(color)

def get_color_rgb_by_index(index):
    if index == 32767:
        index = 0
    return COLOR_INDEX[index % 64]

def get_pos_x(position):
    return re.findall(r'\D+', position)[0] # Match letters

# Start with 1
def get_pos_x_index(position):
    return column_index_from_string(get_pos_x(position))

# Start with 1
def get_pos_y(position):
    return re.findall(r'\d+', position)[0] # Match numbers

def get_pos_y_index(position):
    return int(get_pos_y(position))

def get_left_pos(position, shift=1):
    shift = -shift
    return get_right_pos(position, shift)

def get_right_pos(position, shift=1):
    pos_x = get_pos_x_index(position)
    pos_y = get_pos_y(position)

    pos_x = get_column_letter(pos_x + shift)
    return pos_x + pos_y

def get_up_pos(position, shift=1):
    shift = -shift
    return get_down_pos(position, shift)

def get_down_pos(position, shift=1):
    pos_x = get_pos_x(position)
    pos_y = get_pos_y_index(position)
    pos_y += shift
    if pos_y <= 0:
        raise Exception('Invalid row index %d' % pos_y)

    return pos_x + str(pos_y)

def get_corner_positions(pos_begin, pos_end):
    pos_begin_x_index = get_pos_x_index(pos_begin)
    pos_begin_y_index = get_pos_y_index(pos_begin)
    pos_end_x_index = get_pos_x_index(pos_end)
    pos_end_y_index = get_pos_y_index(pos_end)
    if pos_begin_x_index <= pos_end_x_index:
        begin_x_index = pos_begin_x_index
        end_x_index = pos_end_x_index
    else:
        begin_x_index = pos_end_x_index
        end_x_index = pos_begin_x_index
    if pos_begin_y_index <= pos_end_y_index:
        begin_y_index = pos_begin_y_index
        end_y_index = pos_end_y_index
    else:
        begin_y_index = pos_end_y_index
        end_y_index = pos_begin_y_index
    return begin_x_index, end_x_index, begin_y_index, end_y_index

def get_positions_in_area(pos_begin, pos_end):
    begin_x_index, end_x_index, begin_y_index, end_y_index = get_corner_positions(pos_begin, pos_end)
    pos_list = []
    for x in range(begin_x_index, end_x_index + 1):
        for y in range(begin_y_index, end_y_index + 1):
            pos = get_column_letter(x) + str(y)
            pos_list.append(pos)
    return pos_list

def get_positions_matrix_in_area(pos_begin, pos_end):
    begin_x_index, end_x_index, begin_y_index, end_y_index = get_corner_positions(pos_begin, pos_end)
    pos_matrix = []
    for x in range(begin_x_index, end_x_index + 1):
        pos_list = []
        for y in range(begin_y_index, end_y_index + 1):
            pos = get_column_letter(x) + str(y)
            pos_list.append(get_column_letter(x) + str(y))
        pos_matrix.append(pos_list)
    return pos_matrix

def none_to_zero(value):
    if value is None:
        return 0
    else:
        return value

def empty_to_zero(value):
    if value is None or str(value).rstrip() == '':
        return 0
    else:
        return value


# Copyright (c) 2021 RiceXL

from RiceXL.RiceExcel import RiceExcel
from RiceXL.RiceFont import RiceFont
from RiceXL.RiceBorder import RiceBorder
from RiceXL.RiceAlignment import RiceAlignment
from RiceXL.RicePattern import RicePattern
from RiceXL.RiceUtils import get_color_rgb, get_color_rgb_by_index, get_color_index_by_rgb

__AUTHOR__ = 'SNAKE'
__AUTHOR_EMAIL__ = 'admin@wangbaishi.com'
__LICENSE__ = 'MIT'
__VERSION__ = '0.1.0'
__PYTHON__ = '3.6'


from xlutils.filter import process, XLRDReader, XLWTWriter

def new_copy(wb):
    w = XLWTWriter()
    process(
        XLRDReader(wb, 'unknown.xls'),
        w
        )
    return w.output[0][1], w.style_list



import xlwt
from openpyxl.styles import Font
from openpyxl.styles.colors import COLOR_INDEX
from RiceXL.RiceUtils import get_color_rgb

class RiceFont():
    UNDERLINE_DOUBLE = 'double'
    UNDERLINE_DOUBLE_ACCOUNTING = 'doubleAccounting'
    UNDERLINE_SINGLE = 'single'
    UNDERLINE_SINGLE_ACCOUNTING = 'singleAccounting'

    _name = None
    _color_index = None
    _color = None
    _size = None
    _bold = None
    _italic = None
    _underline = None
    _strike = None

    @property
    def name(self):
        return self._name
    
    @name.setter
    def name(self, name):
        self._name = name
        self.font_xls.name = self._name
        self.font_xlsx.name = self._name

    @property
    def color_index(self):
        return self._color_index
    
    @color_index.setter
    def color_index(self, color_index):
        self._color_index = color_index
        self._color = COLOR_INDEX[self._color_index]
        self.font_xls.colour_index = self._color_index
        self.font_xlsx.color = self._color

    @property
    def color(self):
        return self._color
    
    @color.setter
    def color(self, color):
        self._color = color
        self._color_index = COLOR_INDEX.index(self._color)
        self.font_xls.colour_index = self._color_index
        self.font_xlsx.color = self._color
    
    @property
    def size(self):
        return self._size
    
    @size.setter
    def size(self, size):
        self._size = size
        self.font_xls.height = 20 * self._size
        self.font_xlsx.size = self._size

    @property
    def bold(self):
        return self._bold
    
    @bold.setter
    def bold(self, bold):
        self._bold = bold
        self.font_xls.bold = self._bold
        self.font_xlsx.bold = self._bold
    
    @property
    def italic(self):
        return self._italic
    
    @italic.setter
    def italic(self, italic):
        self._italic = italic
        self.font_xls.italic = self._italic
        self.font_xlsx.italic = self._italic
    
    @property
    def underline(self):
        return self._underline
    
    @underline.setter
    def underline(self, underline):
        if underline == self.UNDERLINE_SINGLE or \
                underline == self.UNDERLINE_SINGLE_ACCOUNTING or \
                underline == self.UNDERLINE_DOUBLE or \
                underline == self.UNDERLINE_DOUBLE_ACCOUNTING:
            self._underline = underline
            self.font_xls.underline = True
            self.font_xlsx.underline = self._underline
        else:
            self._underline = 'none'
            self.font_xls.underline = False
            self.font_xlsx.underline = self._underline

    @property
    def strike(self):
        return self._strike
    
    @strike.setter
    def strike(self, strike):
        self._strike = strike
        self.font_xls.struck_out = self._strike
        self.font_xlsx.strike = self._strike
    
    def init_xls(self, font):
        rice_font = RiceFont()
        rice_font.name = font.name
        rice_font.color_index = font.colour_index % 64 if font.colour_index != 32767 else 0
        rice_font.size = font.height / 20
        rice_font.bold = True if font.bold else False
        rice_font.italic = True if font.italic else False
        rice_font.underline = rice_font.UNDERLINE_SINGLE if font.underline else 'none'
        rice_font.strike = True if font.struck_out else False
        return rice_font
    
    def init_xlsx(self, font):
        rice_font = RiceFont()
        rice_font.name = font.name
        color = get_color_rgb(font.color)
        rice_font.color = color
        rice_font.size = font.size
        rice_font.bold = font.bold
        rice_font.italic = font.italic
        rice_font.underline = font.underline
        rice_font.strike = font.strike
        return rice_font

    def __init__(self):
        self.font_xls = xlwt.Font()
        self.font_xlsx = Font()


import xlwt
from openpyxl.styles import Border, Side
from openpyxl.styles.colors import COLOR_INDEX
from RiceXL.RiceUtils import get_color_rgb, get_color_rgb_by_index

BORDER_STYLE = (
    'none', #0
    'thin', #1
    'medium', #2
    'dashed', #3
    'dotted', #4
    'thick', #5
    'double', #6
    'hair', #7
    'mediumDashed', #8
    'dashDot', #9
    'mediumDashDot', #10
    'dashDotDot', #11
    'mediumDashDotDot', #12
    'slantDashDot' #13
)

class RiceBorder():
    _border_style_left = None
    _border_style_right = None
    _border_style_top = None
    _border_style_bottom = None
    _border_style_all = None
    _border_color_left = None
    _border_color_right = None
    _border_color_top = None
    _border_color_bottom = None
    _border_color_all = None
    _side_left = Side()
    _side_right = Side()
    _side_top = Side()
    _side_bottom = Side()
    _side_all = Side()

    @property
    def border_style_left(self):
        return self._border_style_left
    
    @border_style_left.setter
    def border_style_left(self, border_style):
        self._border_style_left = border_style
        self._border_style_all = None
        self.border_xls.left = BORDER_STYLE.index(self._border_style_left)
        if self.border_xlsx.left is not None:
            self._side_left = self.border_xlsx.left
        else:
            self.border_xlsx.left = self._side_left
        # Side().border_style <==> Side().style
        self._side_left.border_style = self._border_style_left
    
    @property
    def border_style_right(self):
        return self._border_style_right

    @border_style_right.setter
    def border_style_right(self, border_style):
        self._border_style_right = border_style
        self._border_style_all = None
        self.border_xls.right = BORDER_STYLE.index(self._border_style_right)
        if self.border_xlsx.right is not None:
            self._side_right = self.border_xlsx.right
        else:
            self.border_xlsx.right = self._side_right
        self._side_right.border_style = self._border_style_right

    @property
    def border_style_top(self):
        return self._border_style_top

    @border_style_top.setter
    def border_style_top(self, border_style):
        self._border_style_top = border_style
        self._border_style_all = None
        self.border_xls.top = BORDER_STYLE.index(self._border_style_top)
        if self.border_xlsx.top is not None:
            self._side_top = self.border_xlsx.top
        else:
            self.border_xlsx.top = self._side_top
        self._side_top.border_style = self._border_style_top
    
    @property
    def border_style_bottom(self):
        return self._border_style_bottom

    @border_style_bottom.setter
    def border_style_bottom(self, border_style):
        self._border_style_bottom = border_style
        self._border_style_all = None
        self.border_xls.bottom = BORDER_STYLE.index(self._border_style_bottom)
        if self.border_xlsx.bottom is not None:
            self._side_bottom = self.border_xlsx.bottom
        else:
            self.border_xlsx.bottom = self._side_bottom
        self._side_bottom.border_style = self._border_style_bottom
    
    @property
    def border_style_all(self):
        return self._border_style_all

    @border_style_all.setter
    def border_style_all(self, border_style):
        self.border_style_left = border_style
        self.border_style_right = border_style
        self.border_style_top = border_style
        self.border_style_bottom = border_style
        # self._border_style_all must be set in the end,
        # or it would be set as None by the other properties.
        self._border_style_all = border_style
    
    @property
    def border_color_left(self):
        return self._border_color_left

    @border_color_left.setter
    def border_color_left(self, border_color):
        self._border_color_left = border_color
        self._border_color_all = None
        self.border_xls.left_colour = COLOR_INDEX.index(self._border_color_left)
        if self.border_xlsx.left is not None:
            self._side_left = self.border_xlsx.left
        else:
            self.border_xlsx.left = self._side_left
        self._side_left.color = self._border_color_left
    
    @property
    def border_color_right(self):
        return self._border_color_right

    @border_color_right.setter
    def border_color_right(self, border_color):
        self._border_color_right = border_color
        self._border_color_all = None
        self.border_xls.right_colour = COLOR_INDEX.index(self._border_color_right)
        if self.border_xlsx.right is not None:
            self._side_right = self.border_xlsx.right
        else:
            self.border_xlsx.right = self._side_right
        self._side_right.color = self._border_color_right
    
    @property
    def border_color_top(self):
        return self._border_color_top

    @border_color_top.setter
    def border_color_top(self, border_color):
        self._border_color_top = border_color
        self._border_color_all = None
        self.border_xls.top_colour = COLOR_INDEX.index(self._border_color_top)
        if self.border_xlsx.top is not None:
            self._side_top = self.border_xlsx.top
        else:
            self.border_xlsx.top = self._side_top
        self._side_top.color = self._border_color_top
    
    @property
    def border_color_bottom(self):
        return self._border_color_bottom

    @border_color_bottom.setter
    def border_color_bottom(self, border_color):
        self._border_color_bottom = border_color
        self._border_color_all = None
        self.border_xls.bottom_colour = COLOR_INDEX.index(self._border_color_bottom)
        if self.border_xlsx.bottom is not None:
            self._side_bottom = self.border_xlsx.bottom
        else:
            self.border_xlsx.bottom = self._side_bottom
        self._side_bottom.color = self._border_color_bottom
    
    @property
    def border_color_all(self):
        return self._border_color_all
    
    @border_color_all.setter
    def border_color_all(self, border_color):
        self.border_color_left = border_color
        self.border_color_right = border_color
        self.border_color_top = border_color
        self.border_color_bottom = border_color
        # self._border_color_all must be set in the end,
        # or it would be set as None by the other properties.
        self._border_color_all = border_color
    
    def init_xls(self, border):
        rice_border = RiceBorder()
        rice_border.border_style_left = BORDER_STYLE[border.left]
        rice_border.border_style_right = BORDER_STYLE[border.right]
        rice_border.border_style_top = BORDER_STYLE[border.top]
        rice_border.border_style_bottom = BORDER_STYLE[border.bottom]
        rice_border.border_color_left = get_color_rgb_by_index(border.left_colour)
        rice_border.border_color_right = get_color_rgb_by_index(border.right_colour)
        rice_border.border_color_top = get_color_rgb_by_index(border.top_colour)
        rice_border.border_color_bottom = get_color_rgb_by_index(border.bottom_colour)
        return rice_border
    
    def init_xlsx(self, border):
        rice_border = RiceBorder()
        side_left = border.left
        side_right = border.right
        side_top = border.top
        side_bottom = border.bottom
        rice_border.border_style_left = side_left.border_style if side_left.border_style is not None else 'none'
        rice_border.border_style_right = side_right.border_style if side_right.border_style is not None else 'none'
        rice_border.border_style_top = side_top.border_style if side_top.border_style is not None else 'none'
        rice_border.border_style_bottom = side_bottom.border_style if side_bottom.border_style is not None else 'none'
        rice_border.border_color_left = get_color_rgb(side_left.color) if side_left.color is not None else '00000000'
        rice_border.border_color_right = get_color_rgb(side_right.color) if side_right.color is not None else '00000000'
        rice_border.border_color_top = get_color_rgb(side_top.color) if side_top.color is not None else '00000000'
        rice_border.border_color_bottom = get_color_rgb(side_bottom.color) if side_bottom.color is not None else '00000000'
        return rice_border

    def __init__(self):
        self.border_xls = xlwt.Borders()
        self.border_xlsx = Border()


import os
import xlrd
import xlwt
import tempfile
from xlwt.Style import XFStyle
from xlutils.copy import copy
from openpyxl import Workbook
from openpyxl import load_workbook
from openpyxl.utils import get_column_letter
from openpyxl.utils import column_index_from_string
from RiceXL.RiceFont import RiceFont
from RiceXL.RiceBorder import RiceBorder
from RiceXL.RicePattern import RicePattern
from RiceXL.RiceAlignment import RiceAlignment
from RiceXL.xlutils_newcopy import new_copy
from RiceXL.RiceUtils import get_pos_x_index
from RiceXL.RiceUtils import get_pos_y_index
from RiceXL.RiceUtils import get_right_pos
from RiceXL.RiceUtils import get_down_pos
from RiceXL.RiceUtils import get_corner_positions
from RiceXL.RiceUtils import get_positions_in_area

class RiceExcel(object):
    def __init__(self, file_path='Temp.xlsx', writable=False):
        self.excel_path = file_path
        self.writable = writable
        self.sheet_index = None
        self.sheet_name = None
        self.tempfd = None
        self.tempfilename = None

        if self.excel_path.lower().endswith('.xls'):
            self.excel_version = 'xls'
            if os.path.exists(self.excel_path):
                # If formatting_info=False: cannot read styles
                self.workbook = xlrd.open_workbook(self.excel_path, formatting_info=True)
                if writable:
                    # A new workbook for writing
                    self.workbook_write, self.style_list = new_copy(self.workbook)
            elif writable:
                self.workbook_write = xlwt.Workbook(encoding='utf-8')
            else:
                raise Exception('File not found: {}'.format(self.excel_path))
        elif self.excel_path.lower().endswith('.xlsx'):
            self.excel_version = 'xlsx'
            if os.path.exists(self.excel_path):
                self.workbook = load_workbook(filename=self.excel_path, data_only=True)
            elif writable:
                self.workbook = Workbook()
            else:
                raise Exception('File not found: {}'.format(self.excel_path))
        else:
            raise Exception('Unsupported file format!')
    
    def set_sheet_by_index(self, sheet_index):
        self.sheet_index = sheet_index
        self.sheet_name = self.get_sheet_name_by_index(self.sheet_index)
        if self.excel_version == 'xls':
            self.sheet = self.workbook.sheet_by_index(self.sheet_index)
            if self.writable:
                self.sheet_write = self.workbook_write.get_sheet(self.sheet_index)
        elif self.excel_version == 'xlsx':
            self.sheet = self.workbook.worksheets[self.sheet_index]

    def set_sheet_by_name(self, sheet_name):
        self.sheet_name = sheet_name
        self.sheet_index = self.get_sheet_index_by_name(self.sheet_name)
        if self.excel_version == 'xls':
            self.sheet = self.workbook.sheet_by_name(self.sheet_name)
            if self.writable:
                self.sheet_write = self.workbook_write.get_sheet(self.sheet_name)
        elif self.excel_version == 'xlsx':
            self.sheet = self.workbook[self.sheet_name]
    
    def get_sheet_index_by_name(self, sheet_name):
        if self.excel_version == 'xls':
            sheet_index = self.workbook._sheet_names.index(sheet_name)
        elif self.excel_version == 'xlsx':
            sheet_index = self.workbook.sheetnames.index(sheet_name)
        return sheet_index
    
    def get_sheet_name_by_index(self, sheet_index):
        if self.excel_version == 'xls':
            sheet_name = self.workbook._sheet_names[sheet_index]
        elif self.excel_version == 'xlsx':
            sheet_name = self.workbook.sheetnames[sheet_index]
        return sheet_name
    
    def get_sheet(self, writable=False, sheet_name=None, sheet_index=None):
        if sheet_name is not None and sheet_index is not None:
            raise Exception('Cannot set sheet_index & sheet_name at the same time!')
        if writable:
            self.check_writable()
        if sheet_index is not None:
            self.get_sheet_by_index(sheet_index, writable)
        elif sheet_name is not None:
            self.get_sheet_by_name(sheet_name, writable)
        else:
            self.get_current_sheet(writable)
    
    def get_sheet_by_index(self, sheet_index, writable=False):
        if writable:
            self.check_writable()
        if self.excel_version == 'xls':
            if writable:
                return self.workbook_write.get_sheet(sheet_index)
            else:
                return self.workbook.sheet_by_index(sheet_index)
        elif self.excel_version == 'xlsx':
            return self.workbook.worksheets[sheet_index]
    
    def get_sheet_by_name(self, sheet_name, writable=False):
        if writable:
            self.check_writable()
        if self.excel_version == 'xls':
            if writable:
                return self.workbook_write.get_sheet(sheet_name)
            else:
                return self.workbook.sheet_by_name(sheet_name)
        elif self.excel_version == 'xlsx':
            return self.workbook[sheet_name]
    
    def get_current_sheet(self, writable=False):
        self.check_sheet_set()
        if writable:
            self.check_writable()
        if self.excel_version == 'xls':
            if writable:
                return self.sheet_write
            else:
                return self.sheet
        elif self.excel_version == 'xlsx':
            return self.sheet
    
    def get_current_sheet_index(self):
        self.check_sheet_set()
        return self.sheet_index
    
    def get_current_sheet_name(self):
        self.check_sheet_set()
        return self.sheet_name
    
    def get_value(self, position, sheet_index=None, sheet_name=None):
        self.set_sheet_by_index_or_name(sheet_index, sheet_name)
        self.check_sheet_set()
        if self.excel_version == 'xls':
            return self.get_value_xls(position)
        elif self.excel_version == 'xlsx':
            return self.get_value_xlsx(position)
    
    def get_value_xls(self, position):
        self.check_sheet_set()
        pos_x = get_pos_x_index(position)
        pos_y = get_pos_y_index(position)
        pos_x -= 1
        pos_y = int(pos_y) - 1
        
        if pos_y >= self.sheet.nrows or pos_x >= self.sheet.ncols:
            return None
        return self.sheet.cell(pos_y, pos_x).value

    def get_value_xlsx(self, position):
        self.check_sheet_set()
        return self.sheet[position].value

    def set_sheet_by_index_or_name(self, sheet_index, sheet_name):
        if sheet_index is not None and sheet_name is not None:
            raise Exception('Cannot set sheet_index & sheet_name at the same time!')
        if sheet_index is not None:
            self.set_sheet_by_index(sheet_index)
        if sheet_name is not None:
            self.set_sheet_by_name(sheet_name)
    
    def check_sheet_set(self):
        if self.sheet_index is None or self.sheet_name is None:
            raise Exception('Have not set sheet!')

    def create_sheet(self, sheet_name, use_sheet=False):
        self.check_writable()
        if self.excel_version == 'xls':
            self.sheet_write = self.workbook_write.add_sheet(sheet_name, cell_overwrite_ok=False)
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            self.sheet = self.workbook.create_sheet(sheet_name)
        if use_sheet == True:
            self.set_sheet_by_name(sheet_name)
    
    def remove_sheet(self, sheet_name):
        self.check_writable()
        if self.excel_version == 'xls':
            sheet_del = self.get_sheet(True, sheet_name)
            self.workbook_write._Workbook__worksheets.remove(sheet_del)
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            sheet_del = self.get_sheet(True, sheet_name)
            self.sheet = self.workbook.remove(sheet_del)
        if sheet_name == self.sheet_name:
            self.sheet_name = None
            self.sheet_index = None
            self.sheet = None
            self.sheet_write = None

    def get_active_sheet(self):
        if self.excel_version == 'xls':
            if self.writable:
                return self.get_sheet(writable=True, sheet_index=self.workbook_write.active_sheet)
            else:
                workbook_write = copy(self.workbook)
                return workbook_write.get_sheet(workbook_write.active_sheet)
        elif self.excel_version == 'xlsx':
            return self.workbook.active
    
    def get_active_sheet_name(self):
        if self.excel_version == 'xls':
            if self.writable:
                return self.get_sheet_name_by_index(self.workbook_write.active_sheet)
            else:
                workbook_write = copy(self.workbook)
                return self.get_sheet_name_by_index(workbook_write.active_sheet)
        elif self.excel_version == 'xlsx':
            return self.get_sheet_name_by_index(self.workbook._active_sheet_index)

    def set_active_sheet(self, sheet_name):
        self.check_writable()
        if self.excel_version == 'xls':
            self.workbook_write.active_sheet = self.get_sheet_index_by_name(sheet_name)
        elif self.excel_version == 'xlsx':
            self.workbook._active_sheet_index = self.get_sheet_index_by_name(sheet_name)
    
    def set_value(self, position, value, style=None):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            pos_x = get_pos_x_index(position) - 1
            pos_y = get_pos_y_index(position) - 1
            if style is None:
                style = self.get_origin_style_by_index(pos_y, pos_x)
                self.sheet_write.write(pos_y, pos_x, value, style)
            elif style == '':
                self.sheet_write.write(pos_y, pos_x, value)
            else:
                self.sheet_write.write(pos_y, pos_x, value, style)
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            self.sheet[position].value = value
    
    def handle_temp_file(self):
        self.check_writable()
        if self.tempfd is None:
            self.tempfd, self.tempfilename = tempfile.mkstemp(suffix='.xls', prefix='RiceXL')
        self.save(self.tempfilename, True)
        self.workbook = xlrd.open_workbook(self.tempfilename, formatting_info=True)
        temp_workbook_write, self.style_list = new_copy(self.workbook)
        self.set_sheet_by_index(self.sheet_index)
    
    def get_origin_style_by_position(self, position):
        pos_x = get_pos_x_index(position) - 1
        pos_y = get_pos_y_index(position) - 1        
        return self.get_origin_style_by_index(pos_y, pos_x)

    def get_origin_style_by_index(self, pos_y, pos_x):
        self.check_sheet_set()
        try:
            xf_index = self.sheet.cell_xf_index(pos_y, pos_x)
            style = self.style_list[xf_index]
        except Exception:
            style = XFStyle()
        return style
    
    def set_number_format(self, position, format='General'):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            value = self.get_value_xls(position)
            style.num_format_str = format
            self.set_value(position, value, style)
        elif self.excel_version == 'xlsx':
            self.sheet[position].number_format = format
    
    def get_number_format(self, position):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            number_format = style.num_format_str
        elif self.excel_version == 'xlsx':
            number_format = self.sheet[position].number_format
        return number_format
    
    def set_font(self, position, rice_font):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            value = self.get_value_xls(position)
            style.font = rice_font.font_xls
            self.set_value(position, value, style)
        elif self.excel_version == 'xlsx':
            self.sheet[position].font = rice_font.font_xlsx
    
    def get_font(self, position):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            rice_font = RiceFont().init_xls(style.font)
        elif self.excel_version == 'xlsx':
            rice_font = RiceFont().init_xlsx(self.sheet[position].font)
        return rice_font
    
    def set_border(self, position, rice_border):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            value = self.get_value_xls(position)
            style.borders = rice_border.border_xls
            self.set_value(position, value, style)
        elif self.excel_version == 'xlsx':
            self.sheet[position].border = rice_border.border_xlsx
    
    def get_border(self, position):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            rice_border = RiceBorder().init_xls(style.borders)
        elif self.excel_version == 'xlsx':
            rice_border = RiceBorder().init_xlsx(self.sheet[position].border)
        return rice_border
    
    def set_alignment(self, position, rice_alignment):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            value = self.get_value_xls(position)
            style.alignment = rice_alignment.alignment_xls
            self.set_value(position, value, style)
        elif self.excel_version == 'xlsx':
            self.sheet[position].alignment = rice_alignment.alignment_xlsx
    
    def get_alignment(self, position):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            rice_alignment = RiceAlignment().init_xls(style.alignment)
        elif self.excel_version == 'xlsx':
            rice_alignment = RiceAlignment().init_xlsx(self.sheet[position].alignment)
        return rice_alignment
    
    def set_pattern(self, position, rice_pattern):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            value = self.get_value_xls(position)
            style.pattern = rice_pattern.pattern_xls
            self.set_value(position, value, style)
        elif self.excel_version == 'xlsx':
            self.sheet[position].fill = rice_pattern.pattern_xlsx
    
    def get_pattern(self, position):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            style = self.get_origin_style_by_position(position)
            rice_pattern = RicePattern().init_xls(style.pattern)
        elif self.excel_version == 'xlsx':
            rice_pattern = RicePattern().init_xlsx(self.sheet[position].fill)
        return rice_pattern
    
    def set_height(self, row, height):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            row -= 1
            self.sheet_write.row(row).height_mismatch = True
            self.sheet_write.row(row).height = int(20 * height)
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            self.sheet.row_dimensions[row].height = height
    
    def get_height(self, row):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            row -= 1
            if row >= len(self.sheet.rowinfo_map):
                height = 20 * 12.75
            else:
                height = self.sheet.rowinfo_map[row].height
            height /= 20
        elif self.excel_version == 'xlsx':
            height = self.sheet.row_dimensions[row].height
            if height is None:
                height = 13.5
        return height
    
    def set_width(self, col, width):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            col_index = column_index_from_string(col) - 1
            self.sheet_write.col(col_index).width = int(256 * width)
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            self.sheet.column_dimensions[col].width = width
    
    def get_width(self, col):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            col_index = column_index_from_string(col) - 1
            if self.sheet.colinfo_map.get(col_index) is not None:
                width = round(self.sheet.colinfo_map[col_index].width / 256, 1)
            else:
                width = 9.1
        elif self.excel_version == 'xlsx':
            width = self.sheet.column_dimensions[col].width
        return width
    
    def hide_row(self, row):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            self.set_height(row, 0)
        elif self.excel_version == 'xlsx':
            self.sheet.row_dimensions[row].hidden = True
    
    def unhide_row(self, row):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            height = self.get_height(row)
            if height == 0:
                height = 12.75
            self.set_height(row, height)
        elif self.excel_version == 'xlsx':
            self.sheet.row_dimensions[row].hidden = False
    
    def is_row_hidden(self, row):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            height = self.get_height(row)
            if height == 0:
                hidden = True
            else:
                hidden = False
        elif self.excel_version == 'xlsx':
            hidden = self.sheet.row_dimensions[row].hidden
        return hidden
    
    def hide_col(self, col):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            self.set_width(col, 0)
        elif self.excel_version == 'xlsx':
            self.sheet.column_dimensions[col].hidden = True
    
    def unhide_col(self, col):
        self.check_sheet_set()
        self.check_writable()
        if self.excel_version == 'xls':
            width = self.get_width(col)
            if width == 0:
                width = 9.1
            self.set_width(col, width)
        elif self.excel_version == 'xlsx':
            self.sheet.column_dimensions[col].hidden = False
    
    def is_col_hidden(self, col):
        self.check_sheet_set()
        if self.excel_version == 'xls':
            width = self.get_width(col)
            if width == 0:
                hidden = True
            else:
                hidden = False
        elif self.excel_version == 'xlsx':
            hidden = self.sheet.column_dimensions[col].hidden
        return hidden
    
    def merge_cells(self, pos_begin, pos_end):
        self.check_sheet_set()
        self.check_writable()
        begin_x_index, end_x_index, begin_y_index, end_y_index = get_corner_positions(pos_begin, pos_end)
        if self.excel_version == 'xls':
            pos_first = get_column_letter(begin_x_index) + str(begin_y_index)
            value = self.get_value_xls(pos_first)
            style = self.get_origin_style_by_position(pos_first)
            begin_x_index -= 1
            end_x_index -= 1
            begin_y_index -= 1
            end_y_index -= 1
            self.sheet_write.write_merge(begin_y_index, end_y_index, begin_x_index, end_x_index, value, style)
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            self.sheet.merge_cells(start_row=begin_y_index, start_column=begin_x_index, end_row=end_y_index, end_column=end_x_index)
    
    def unmerge_cells(self, pos_begin, pos_end):
        self.check_sheet_set()
        self.check_writable()
        begin_x_index, end_x_index, begin_y_index, end_y_index = get_corner_positions(pos_begin, pos_end)
        if self.excel_version == 'xls':
            begin_x_index -= 1
            end_x_index -= 1
            begin_y_index -= 1
            end_y_index -= 1
            self.sheet_write.merged_ranges.remove((begin_y_index, end_y_index, begin_x_index, end_x_index))
            self.handle_temp_file()
        elif self.excel_version == 'xlsx':
            self.sheet.unmerge_cells(start_row=begin_y_index, start_column=begin_x_index, end_row=end_y_index, end_column=end_x_index)

    def save(self, new_file_path=None, temp_flag=False):
        self.check_writable()
        if new_file_path is None:
            new_file_path = self.excel_path
        if self.excel_version == 'xls':
            self.workbook_write.save(new_file_path)
            if not temp_flag:
                self.workbook_write = copy(self.workbook)
        elif self.excel_version == 'xlsx':
            self.workbook.save(new_file_path)

    def clean_up(self):
        if self.tempfd is not None:
            os.close(self.tempfd)
            os.unlink(self.tempfilename)
            self.tempfd = None
            self.tempfilename = None

    def check_writable(self):
        if not self.writable:
            raise Exception('The file is not writable!')
    
    def sum_by_positions(self, *positions):
        self.check_sheet_set()
        sum = 0
        for pos in positions:
            sum += float(self.get_value(pos))
        return sum
    
    def sum_in_area(self, pos_begin, pos_end):
        self.check_sheet_set()
        pos_list = get_positions_in_area(pos_begin, pos_end)
        sum = 0
        for pos in pos_list:
            sum += float(self.get_value(pos))
        return sum
    
    def get_values_matrix_in_area(self, pos_begin, pos_end):
        self.check_sheet_set()
        begin_x_index, end_x_index, begin_y_index, end_y_index = get_corner_positions(pos_begin, pos_end)
        value_matrix = []
        for x in range(begin_x_index, end_x_index + 1):
            value_list = []
            for y in range(begin_y_index, end_y_index + 1):
                pos = get_column_letter(x) + str(y)
                value = self.get_value(pos)
                value_list.append(value)
            value_matrix.append(value_list)
        return value_matrix
    
    def set_value_by_area(self, pos_begin, value_matrix):
        self.check_sheet_set()
        self.check_writable()
        pos_cursor = pos_begin
        for value_list in value_matrix:
            pos = pos_cursor
            for value in value_list:
                self.set_value(pos, value)
                pos = get_down_pos(pos)
            pos_cursor = get_right_pos(pos_cursor)


import xlwt
from openpyxl.styles import Alignment

HORIZONTAL = (
    'general', #0
    'left', #1
    'center', #2
    'right', #3
    'fill', #4
    'justify', #5
    'centerContinuous', #6
    'distributed' #7
)

VERTICAL = (
    'top', #0
    'center', #1
    'bottom', #2
    'justify', #3
    'distributed' #4
)

class RiceAlignment():
    _horizontal = None
    _vertical = None
    _wrap = None

    @property
    def horizontal(self):
        return self._horizontal
    
    @horizontal.setter
    def horizontal(self, horizontal):
        self._horizontal = horizontal
        self.alignment_xls.horz = HORIZONTAL.index(horizontal)
        self.alignment_xlsx.horizontal = horizontal
    
    @property
    def vertical(self):
        return self._vertical
    
    @vertical.setter
    def vertical(self, vertical):
        self._vertical = vertical
        self.alignment_xls.vert = VERTICAL.index(vertical)
        self.alignment_xlsx.vertical = vertical
    
    @property
    def wrap(self):
        return self._wrap
    
    @wrap.setter
    def wrap(self, wrap):
        self._wrap = wrap
        self.alignment_xls.wrap = 1 if self._wrap else 0
        self.alignment_xlsx.wrap_text = wrap
    
    def init_xls(self, alignment):
        rice_alignment = RiceAlignment()
        rice_alignment.horizontal = HORIZONTAL[alignment.horz]
        rice_alignment.vertical = VERTICAL[alignment.vert]
        rice_alignment.wrap = True if alignment.wrap else False
        return rice_alignment
    
    def init_xlsx(self, alignment):
        rice_alignment = RiceAlignment()
        rice_alignment.horizontal = alignment.horizontal if alignment.horizontal is not None else 'general'
        rice_alignment.vertical = alignment.vertical if alignment.vertical is not None else 'top'
        rice_alignment.wrap = alignment.wrap_text if alignment.wrap_text is not None else False
        return rice_alignment

    def __init__(self):
        self.alignment_xls = xlwt.Alignment()
        self.alignment_xlsx = Alignment()